from django_apiview import extras
import re
import uuid
import logging

from django.apps import apps
from django.db.models import QuerySet
from django.db.models.signals import post_save
from django.db.models.signals import post_delete
from django_redis import get_redis_connection

from ..views import cache as decorator_cache

logger = logging.getLogger(__name__)

class CacheCleanerBase(object):

    def __init__(self, key, model, cache_connection=None, **kwargs):
        self.key = key
        self.model = isinstance(model, str) and apps.get_model(model) or model
        self.cache_connection = cache_connection or get_redis_connection()
        self.kwargs = kwargs
        self.setup()
    
    def setup(self):
        logger.info("Cache setup cleaner, key={0}, model={1}, cleaner={2}".format(self.key, self.model, self.__class__.__name__))
        post_save.connect(self.clean, sender=self.model)
        post_delete.connect(self.clean, sender=self.model)
        return self

    def clean(self, sender, instance, **kwargs):
        try:
            self._clean(sender, instance, **kwargs)
        except Exception:
            logger.exception("Cache do clean failed.")

    def _clean(self, sender, instance, **kwargs):
        key_patterns = self.get_clean_keys(instance)
        if key_patterns:
            logger.info("clean cache: key_patterns={0}".format(key_patterns))
            self.delete_key_patterns(key_patterns)
        else:
            logger.info("clean cache: none")
        return self

    def delete_key_patterns(self, key_patterns):
        keys = []
        for key_pattern in key_patterns:
            if "*" in key_pattern:
                keys += self.cache_connection.keys(key_pattern)
            else:
                keys.append(key_pattern)
        keys = list(set(keys))
        logger.info("{} keys deleted".format(len(keys)))
        if keys:
            self.cache_connection.delete(*keys)
    
    def get_clean_keys(self, instance):
        raise NotImplementedError()

class SimpleCacheCleaner(CacheCleanerBase):
    def get_clean_keys(self, instance):
        return [
            re.sub('\{[^\}]*\}', '*', self.key),
        ]

class CleanByModelFields(CacheCleanerBase):
    fields = []
    extra = {}

    def get_clean_keys(self, instance):
        info = self.get_fields_mapping(instance)
        return [
            self.key.format(**info, **self.extra)
        ]
    
    def get_fields_mapping(self, instance):
        info = {}
        for field in self.fields:
            if isinstance(field, (list, tuple)):
                model_field = field[1]
                field = field[0]
            else:
                model_field = field
            info[field] = getattr(instance, model_field, None)
        return info

    @classmethod
    def make(cls, *fields, **kwargs):
        suffix = str(uuid.uuid4()).replace("-", "")
        cleaner_class = type("CleanByModelFields_" + suffix, (cls,), {"fields": fields, "extra": kwargs})
        return cleaner_class

class CleanByRelatedModelFields(CacheCleanerBase):
    related = None
    fields = []
    extra = {}

    def get_clean_keys(self, instance):
        infos = self.get_related_objects_fields_mapping(instance)
        keys = []
        for info in infos:
            keys.append(self.key.format(**info, **self.extra))
        return keys

    def get_related_objects_fields_mapping(self, instance):
        infos = []
        related_instances = getattr(instance, self.related)
        if isinstance(related_instances, QuerySet):
            related_instances = related_instances.all()
        else:
            related_instances = [related_instances]
        for related_instance in related_instances:
            info = {}
            for field in self.fields:
                if isinstance(field, (list, tuple)):
                    model_field = field[1]
                    field = field[0]
                else:
                    model_field = field
                info[field] = getattr(related_instance, model_field, None)
            infos.append(info)
        return infos

    @classmethod
    def make(cls, related, *fields, **kwargs):
        suffix = str(uuid.uuid4()).replace("-", "")
        cleaner_class = type("CleanByRelatedModelFields_" + suffix, (cls,), {"related": related, "fields": fields, "extra": kwargs})
        return cleaner_class

class CleanByImplicitRelatedModelFields(CacheCleanerBase):
    related_model = None
    related_fields = []
    fields = []
    extra = {}

    def get_clean_keys(self, instance):
        infos = self.get_related_objects_fields_mapping(instance)
        keys = []
        for info in infos:
            keys.append(self.key.format(**info, **self.extra))
        return keys
    
    def get_related_objects_fields_mapping(self, instance):
        infos = []
        params = {}
        for related_field in self.related_fields:
            if isinstance(related_field, (list, tuple)):
                to_field = related_field[1]
                related_field = related_field[0]
            else:
                to_field = related_field
            params[to_field] = getattr(instance, related_field, None)
        related_instances = self.related_model.objects.filter(**params).all()
        for related_instance in related_instances:
            info = {}
            for field in self.fields:
                if isinstance(field, (list, tuple)):
                    model_field = field[1]
                    field = field[0]
                else:
                    model_field = field
                info[field] = getattr(related_instance, model_field, None)
            infos.append(info)
        return infos

    @classmethod
    def make(cls, related_model, related_fields, *fields, **kwargs):
        if isinstance(related_fields, str):
            related_fields = [related_fields]
        suffix = str(uuid.uuid4()).replace("-", "")
        cleaner_class = type("CleanByImplicitRelatedModelFields_" + suffix, (cls,), {"related_model": related_model, "related_fields": related_fields, "fields": fields, "extra": kwargs})
        return cleaner_class

CACHE_CLEANERS = {}

def _get_model_name(model):
    if isinstance(model, str):
        return model
    return ".".join([model._meta.app_label, model._meta.model_name])

def setup_cleaner(key, model, cleaner=None):
    model_name = _get_model_name(model)
    cleaner = cleaner or SimpleCacheCleaner
    cleaner_key = ":".join([key, model_name])
    if not cleaner_key in CACHE_CLEANERS:
        CACHE_CLEANERS[cleaner_key] = cleaner(key, model)

def auto_discover_caches():
    from django.urls import reverse
    from django.urls import NoReverseMatch
    from django.urls import Resolver404
    try:
        reverse("none-exist-view")
    except NoReverseMatch:
        pass
    except Resolver404:
        pass
    except Exception:
        logger.exception("auto_discover_caches failed on load url configs...")
    return decorator_cache.related_models_mapping

def auto_setup_cleaners():
    logger.info("Cache auto setup cleaners...")
    related_models_mapping = auto_discover_caches()
    for entry_point, info in related_models_mapping.items():
        for model in info.get("related_models", []):
            if isinstance(model, (list, tuple)):
                cleaner = model[1]
                model = model[0]
            else:
                cleaner = None
            key = info.get("key", None)
            if key and model:
                setup_cleaner(key, model, cleaner)
    logger.info("Cache auto setup cleaners done!")
    return related_models_mapping, CACHE_CLEANERS


import logging
import functools
import json
import uuid

import bizerror
from fastutils import rsautils
from fastutils import strutils
from fastutils import funcutils
from fastutils import typingutils
from fastutils import cipherutils

from django.http import JsonResponse
from django.conf import settings

from .utils import SimpleJsonEncoder
from .utils import get_request_data
from .pack import SimpleJsonResultPacker
from .pack import SafeJsonResultPacker

logger = logging.getLogger(__name__)

simple_result_packer = SimpleJsonResultPacker()

def setup_result_packer(packer):
    global simple_result_packer
    simple_result_packer = packer


class ApiviewDecorator(object):
    """Decorator base class.
    """

    def __call__(self, func):
        def wrapper(_django_apiview_request):
            return self.main(_django_apiview_request, func)
        wrapper._apiview_decorator = True
        return functools.wraps(func)(wrapper)

    def main(self, _django_apiview_request, func):
        return _django_apiview_request.process(func)


class ApiviewRequest(object):
    """Process class of apiview.
    """
    
    def __init__(self, request, defaults, **kwargs):
        self._apiview_decorator = False
        self.request = request
        self.kwargs = kwargs
        self.data = {}
        self.data.update(defaults)
        self.data.update(get_request_data(self.request, self.kwargs))
        self.data["_django_apiview_request"] = self

    def process(self, func):
        if hasattr(func, "_apiview_decorator") and getattr(func, "_apiview_decorator"):
            return func(_django_apiview_request=self)
        else:
            return funcutils.call_with_inject(func, self.data)

class Apiview(object):
    def __init__(self, packer, preload_plugins=None, extra_parameters=None, **kwargs):
        self.packer = packer
        self.extra_parameters = extra_parameters or {}
        self.preload_plugins = preload_plugins or []
        self.apiview_init_kwargs = kwargs

    def set_packer(self, packer):
        self.packer = packer

    def add_preload_plugin(self, plugin):
        self.preload_plugins.append(plugin)

    def put_extra_parameter(self, key, value):
        self.extra_parameters[key] = value

    def put_extra_parameters(self, params):
        self.extra_parameters.update(params)

    def __call__(self, func):
        """Turn the view function into apiview function. Must use as the first decorator.
        """
        def wrapper(request, **kwargs):
            target_func = func
            defaults = {}
            defaults.update(self.extra_parameters)
            defaults.update(funcutils.get_default_values(func))
            _django_apiview_request = ApiviewRequest(request, defaults, **kwargs)
            package = {}
            try:
                for plugin in reversed(self.preload_plugins):
                    target_func = plugin(target_func)
                result = _django_apiview_request.process(target_func)
                package = self.packer.pack_result(result, **_django_apiview_request.data)
            except Exception as error:
                logger.exception("apiview process failed: {}".format(str(error)))
                if not isinstance(error, bizerror.BizErrorBase):
                    error = bizerror.BizError(error)
                package = self.packer.pack_error(error, **_django_apiview_request.data)
            return JsonResponse(package, encoder=SimpleJsonEncoder, json_dumps_params={"ensure_ascii": False, "allow_nan": True, "sort_keys": True})
        wrapper.csrf_exempt = True
        return functools.wraps(func)(wrapper)

apiview = Apiview(simple_result_packer)

class requires(ApiviewDecorator):
    """Make sure the parameters are given.
    """
    def __init__(self, *parameter_names):
        self.parameter_names = parameter_names

    def main(self, _django_apiview_request, func):
        missing_names = []
        for name in self.parameter_names:
            if isinstance(name, str):
                if not name in _django_apiview_request.data:
                    missing_names.append(name)
            elif isinstance(name, (list, set, tuple)):
                flag = False
                for element in name:
                    if _django_apiview_request.data.get(element, None):
                        flag = True
                        break
                if not flag:
                    missing_names.append("({0})".format(" or ".join(name)))
        if missing_names:
            raise bizerror.MissingParameter(parameter=",".join(missing_names))
        return super().main(_django_apiview_request, func)

class choices(ApiviewDecorator):
    """Make sure field's value is in the choices.
    """
    def __init__(self, field, choices, annotation=None, allow_none=False):
        self.field = field
        self.choices = choices
        self.annotation = annotation
        self.allow_none = allow_none

    def main(self, _django_apiview_request, func):
        if callable(self.choices):
            params = funcutils.get_inject_params(self.choices, _django_apiview_request.data)
            values = self.choices(**params)
        else:
            values = self.choices
        value = _django_apiview_request.data.get(self.field, None)
        if self.annotation:
            value = typingutils.smart_cast(self.annotation, value)
        if (self.allow_none and value is None) or (value in self.choices):
            return super().main(_django_apiview_request, func)
        else:
            raise bizerror.BadParameter("field {0}'s value '{1}' is not in choices {2}.".format(self.field, value, values))

class between(ApiviewDecorator):
    """Make sure field's numeric value is in range of (min, max).
    """

    def __init__(self, field, min, max, include_min=True, include_max=True, annotation=typingutils.Number, allow_none=False):
        self.field = field
        self.min = min
        self.max = max
        self.include_min = include_min
        self.include_max = include_max
        self.annotation = annotation
        self.allow_none = allow_none
    
    def main(self, _django_apiview_request, func):
        field = self.field
        min = self.min
        max = self.max
        include_min = self.include_min
        include_max = self.include_max
        annotation = self.annotation
        allow_none = self.allow_none

        if callable(min):
            params = funcutils.get_inject_params(min, _django_apiview_request.data)
            min_value = min(**params)
        else:
            min_value = min
        if callable(max):
            params = funcutils.get_inject_params(max, _django_apiview_request.data)
            max_value = max(**params)
        else:
            max_value = max
        value = _django_apiview_request.data.get(field, None)
        value = typingutils.smart_cast(typingutils.Number, value)
        if (allow_none and value is None) or ((include_min and min_value <= value or min_value < value) and (include_max and max_value >= value or max_value > value)):
            return super().main(_django_apiview_request, func)
        else:
            raise bizerror.BadParameter("field {0}'s value '{1}' is not in range of {2}{3}, {4}{5}.".format(
                field, value,
                include_min and "[" or "(", 
                min_value, max_value,
                include_max and "]" or ")",
                ))

class rsa_decrypt(ApiviewDecorator):
    """Do rsa-decrypt to the given field with private_key.
    """

    def __init__(self, field, private_key):
        self.field = field
        self.private_key = private_key
    
    def main(self, _django_apiview_request, func):
        if self.field in _django_apiview_request.data:
            field_value = _django_apiview_request.data[self.field]
            field_data = rsautils.smart_get_binary_data(field_value)
            plain_data = rsautils.decrypt(field_data, self.private_key)
            plain_text = plain_data.decode("utf-8")
            _django_apiview_request.data[self.field] = plain_text
        return super().main(_django_apiview_request, func)


class meta_variable(ApiviewDecorator):
    """Get variable from meta.
    """
    def __init__(self, variable_name, meta_name):
        self.variable_name = variable_name
        self.meta_name = meta_name
    
    def main(self, _django_apiview_request, func):
        value = _django_apiview_request.request.META.get(self.meta_name, None)
        _django_apiview_request.data[self.variable_name] = value
        return super().main(_django_apiview_request, func)

class cache(ApiviewDecorator):

    related_models_mapping = {}

    @classmethod
    def get_entry_points(cls, app_label=None):
        if not app_label:
            eps = list(cls.related_models_mapping.values())
            eps.sort(key=lambda x:x["entry_point"])
            return eps
        else:
            eps = []
            for ep in list(cls.related_models_mapping.keys()):
                if app_label == ep.split(".")[0]:
                    eps.append(cls.related_models_mapping[ep])
            eps.sort(key=lambda x:x["entry_point"])
            return eps

    def keep_related_models_mapping(self):
        entry_point = id(self)
        self.related_models_mapping[entry_point] = {
            "entry_point": entry_point,
            "related_models": [] + self.related_models,
            "key": self.key,
        }
        logger.info("Cache entry discovered: {0}".format(self.related_models_mapping[entry_point]))

    def __init__(self, key, expire=None, cache_name="default", get_from_cache=None, set_to_cache=True, disable_get_from_cache_header=None, batch_mode=False, ignore_cache_errors=True, related_models=None):
        self.key = key
        self.expire = expire or getattr(settings, "DJANGO_APIVIEW_DEFAULT_CACHE_EXPIRE", None)
        self.expire = self.expire and int(self.expire) or None
        self.cache_name = cache_name
        self.get_from_cache = get_from_cache
        self.set_to_cache = set_to_cache
        self.disable_get_from_cache_header = disable_get_from_cache_header or getattr(settings, "DJANGO_APIVIEW_DISABLE_CACHE_HEADER_NAME", "HTTP_DISABLE_CACHE")
        self.batch_mode = batch_mode
        self.ignore_cache_errors = ignore_cache_errors
        self.related_models = related_models or []
        self.keep_related_models_mapping()

    def main(self, _django_apiview_request, func):
        # Import here, so that we don't need django-redis by default.
        # Only if you use cache, then you need pip install django-redis
        from django_redis import get_redis_connection

        key_template = self.key
        expire = self.expire
        cache_name = self.cache_name
        get_from_cache = self.get_from_cache
        set_to_cache = self.set_to_cache
        disable_get_from_cache_header = self.disable_get_from_cache_header
        batch_mode = self.batch_mode
        ignore_cache_errors = self.ignore_cache_errors

        if get_from_cache == False:
            get_from_cache_final = False
        else:
            if _django_apiview_request.request.META.get(disable_get_from_cache_header, "false") == "true":
                get_from_cache_final = False
            else:
                get_from_cache_final = True

        set_to_cache_final = set_to_cache
        use_cache = get_from_cache_final or set_to_cache_final

        cache = None
        if use_cache:
            try:
                cache = get_redis_connection(cache_name)
            except Exception as error:
                logger.exception("get redis connection failed @cache, error_message={0}".format(str(error)))
                if ignore_cache_errors:
                    cache = None
                else:
                    raise

        if cache and get_from_cache:
            try:
                key = strutils.format_with_mapping(
                    key_template,
                    funcutils.chain(strutils.none_to_empty_string, strutils.strip_string),
                    **_django_apiview_request.data,
                    )
                result_text = cache.get(key)
                if not result_text is None:
                    return json.loads(result_text)
            except Exception as error:
                logger.exception("query redis failed @cache, error_message={0}".format(str(error)))
                if not ignore_cache_errors:
                    raise

        result = super().main(_django_apiview_request, func)

        def _set_cache(result, **kwargs):
            if isinstance(result, dict):
                key = strutils.format_with_mapping(
                    key_template,
                    funcutils.chain(strutils.none_to_empty_string, strutils.strip_string),
                    **_django_apiview_request.data, **result, **kwargs
                    )
            else:
                key = strutils.format_with_mapping(
                    key_template,
                    funcutils.chain(strutils.none_to_empty_string, strutils.strip_string),
                    **_django_apiview_request.data, **kwargs,
                    )
            result_text = json.dumps(result, cls=SimpleJsonEncoder, allow_nan=True, sort_keys=True)
            cache.set(key, result_text)
            if expire:
                cache.expire(key, expire) # Old version redis don't support add ttl while doing set key value, so do set ttl separately
        
        if cache and set_to_cache_final:
            try:
                if batch_mode:
                    if isinstance(result, list):
                        for element in result:
                            _set_cache(element)
                    elif isinstance(result, dict):
                        for key, value in result.items():
                            _set_cache(value, __key=key, __value=value)
                else:
                    _set_cache(result)
            except Exception as error:
                logger.exception("write redis failed @cache, error_message={0}".format(str(error)))
                if not ignore_cache_errors:
                    raise

        return result

class decode_encrypted_data(ApiviewDecorator):

    def __init__(self, result_encoder=cipherutils.SafeBase64Encoder(), privatekey=None, server_rsa_privatekey_filedname="RSA_PRIVATEKEY", encrypted_password_fieldname="encryptedPassword", encrypted_data_fieldname="encryptedData"):
        self.result_encoder = result_encoder
        self.server_rsa_privatekey_filedname = server_rsa_privatekey_filedname
        self.privatekey = privatekey
        self.encrypted_password_fieldname = encrypted_password_fieldname
        self.encrypted_data_fieldname = encrypted_data_fieldname

    def main(self, _django_apiview_request, func):
        privatekey = self.privatekey or getattr(settings, self.server_rsa_privatekey_filedname, None)
        encrypted_password = _django_apiview_request.data.get(self.encrypted_password_fieldname, "")
        encrypted_data = _django_apiview_request.data.get(self.encrypted_data_fieldname, "")
        if privatekey and encrypted_password and encrypted_data:
            password = rsautils.decrypt(encrypted_password, privatekey)
            cipher = cipherutils.AesCipher(password=password, result_encoder=self.result_encoder)
            data_text = cipher.decrypt(encrypted_data)
            data = json.loads(data_text)
            _django_apiview_request.data.update(data)
        return super().main(_django_apiview_request, func)

def safe_apiview(packer_class=SafeJsonResultPacker, **kwargs):
    return Apiview(
        packer=funcutils.call_with_inject(packer_class, kwargs),
        preload_plugins = [
            funcutils.call_with_inject(decode_encrypted_data, kwargs),
            ],
        extra_parameters=kwargs,
        )


DEFAULT_RANDOM_DJANGO_APIVIEW_ACLKEY = str(uuid.uuid4())
logger.info("Default random DJANGO_APIVIEW_ACLKEY={0}".format(DEFAULT_RANDOM_DJANGO_APIVIEW_ACLKEY))

class check_aclkey(ApiviewDecorator):

    def __init__(self, aclkey=None, aclkey_field_name="aclkey"):
        self.aclkey_field_name = aclkey_field_name
        self.aclkey = aclkey or getattr(settings, "DJANGO_APIVIEW_ACLKEY", DEFAULT_RANDOM_DJANGO_APIVIEW_ACLKEY)
    
    def main(self, _django_apiview_request, func):
        if not self.aclkey_field_name in _django_apiview_request.data:
            raise bizerror.MissingParameter(parameter=self.aclkey_field_name)
        aclkey = _django_apiview_request.data[self.aclkey_field_name]
        if not self._check_aclkey(aclkey, _django_apiview_request):
            raise bizerror.AppAuthFailed()
        return super().main(_django_apiview_request, func)

    def _check_aclkey(self, aclkey, _django_apiview_request):
        if callable(self.aclkey):
            aclkey = funcutils.call_with_inject(self.aclkey, _django_apiview_request.data)
        return self.aclkey == aclkey


class string_length_limit(ApiviewDecorator):
    """Check string parameter length.
    """
    def __init__(self, field, max_length, min_length=0, string_too_short_error_message=None, string_too_long_error_message=None):
        self.field = field
        self.max_length = max_length
        self.min_length = min_length
        self.string_too_short_error_message = string_too_short_error_message
        self.string_too_long_error_message = string_too_long_error_message
    
    def main(self, _django_apiview_request, func):
        value = _django_apiview_request.data.get(self.field, None)
        if not value is None:
            value_length = len(value)
            if value_length < self.min_length:
                raise bizerror.StringTooShort(message=self.string_too_short_error_message, min_length=self.min_length)
            if value_length > self.max_length:
                raise bizerror.StringTooLong(message=self.string_too_long_error_message, max_length=self.max_length)
        return super().main(_django_apiview_request, func)






import json

import bizerror
from fastutils import jsonutils
from fastutils import strutils
from fastutils import cipherutils
from fastutils import rsautils



class AbstractResultPacker(object):
    
    def pack_result(self, result, **kwargs):
        raise NotImplementedError()

    def pack_error(self, error, **kwargs):
        raise NotImplementedError()

    def unpack(self, data, **kwargs):
        raise NotImplementedError()



class SimpleJsonResultPacker(AbstractResultPacker):

    def pack_result(self, result, **kwargs):
        return {
            "success": True,
            "result": result,
        }

    def pack_error(self, error, **kwargs):
        return {
            "success": False,
            "error": error,
        }

    def unpack(self, data, **kwargs):
        if not data:
            raise bizerror.BadResponseContent(content=data)
        if isinstance(data, (str, bytes)):
            try:
                data = json.loads(data)
            except Exception:
                raise bizerror.ParseJsonError(text=data)
        success = data.get("success", None)
        if success == True and "result" in data:
            return data["result"]
        elif success == None and "result" in data:
            return data["result"]
        if "error" in data:
            raise bizerror.BizError(data["error"]["message"], data["error"]["code"])
        else:
            raise bizerror.BadResponseContent(content=data)

class SafeJsonResultPacker(SimpleJsonResultPacker):
    
    def __init__(self, result_encoder=cipherutils.SafeBase64Encoder(), password_length=32, client_id_fieldname="clientId", encrypted_password_fieldname="encryptedPassword", encrypted_data_fieldname="encryptedData"):
        self.password_length = password_length
        self.encrypted_password_fieldname = encrypted_password_fieldname
        self.encrypted_data_fieldname = encrypted_data_fieldname
        self.result_encoder = result_encoder
        self.client_id_fieldname = client_id_fieldname

    def pack_result(self, result, **kwargs):
        result = super().pack_result(result, **kwargs)
        return self.encrypt_data(result, **kwargs)

    def pack_error(self, error, **kwargs):
        error =  super().pack_error(error, **kwargs)
        return self.encrypt_data(error, **kwargs)

    def encrypt_data(self, data, **kwargs):
        # get client rsa publickey
        get_client_rsa_publickey = kwargs["get_client_rsa_publickey"]
        client_id = kwargs[self.client_id_fieldname]
        client_rsa_publickey = get_client_rsa_publickey(client_id)
        # do data encrypt
        result_text = jsonutils.simple_json_dumps(data)
        password = strutils.random_string(self.password_length)
        result_cipher = cipherutils.AesCipher(password=password, result_encoder=self.result_encoder)
        result_data = result_cipher.encrypt(result_text.encode("utf-8"))
        encrypted_password = rsautils.encrypt(password.encode(), client_rsa_publickey)
        return {
            self.encrypted_password_fieldname: encrypted_password,
            self.encrypted_data_fieldname: result_data,
        }
 
    def unpack(self, data, **kwargs):
        if not data:
            raise bizerror.BadResponseContent(content=data)
        if isinstance(data, (str, bytes)):
            try:
                data = json.loads(data)
            except Exception:
                raise bizerror.ParseJsonError(text=data)
        client_rsa_privatekey = kwargs["client_rsa_privatekey"]
        encrypted_password = data[self.encrypted_password_fieldname]
        encrypted_data = data[self.encrypted_data_fieldname]
        password = rsautils.decrypt(encrypted_password, client_rsa_privatekey)
        result_cipher = cipherutils.AesCipher(password=password, result_encoder=self.result_encoder)
        data_json = result_cipher.decrypt(encrypted_data)
        return super().unpack(data_json, **kwargs)



import requests

from .pack import SimpleJsonResultPacker


class ApiviewClient(object):

    def __init__(self, packer=None):
        self.packer = packer or SimpleJsonResultPacker()

    def get(self, url, params=None):
        params = params or {}
        response = requests.get(url, params=params)
        return self.packer.unpack(response.content)

    def post(self, url, params=None, data=None):
        params = params or {}
        data = data or {}
        response = requests.post(url, params=params, json=data)
        return self.packer.unpack(response.content)




import json

from fastutils.jsonutils import SimpleJsonEncoder
from fastutils.cacheutils import get_cached_value

from django.core.serializers.json import DjangoJSONEncoder as DjangoJSONEncoderBase
from django.forms.models import model_to_dict


def _get_request_data(request, extra_view_parameters):
    data = {"_request": request}
    data.update(extra_view_parameters)
    for name, _ in request.GET.items():
        value = request.GET.getlist(name)
        if isinstance(value, (list, tuple, set)) and len(value) == 1:
            data[name] = value[0]
        else:
            data[name] = value
    for name, _ in request.POST.items():
        value = request.POST.getlist(name)
        if isinstance(value, (list, tuple, set)) and len(value) == 1:
            data[name] = value[0]
        else:
            data[name] = value
    if request.body:
        try:
            payload = json.loads(request.body)
            data["_form"] = payload
            data.update(payload)
        except:
            pass
    for name, fobj in request.FILES.items():
        data[name] = fobj
    return data

def get_request_data(request, extra_view_parameters):
    return get_cached_value(request, "_django_apiview_request_data", _get_request_data, request, extra_view_parameters)



from pprint import pprint

import djclick as click
from django_apiview.extras.cache_admin import auto_setup_cleaners


@click.command()
def list_cache_entries():
    related_models_mapping, cache_cleaners = auto_setup_cleaners()

    print("Cache related models mapping:")
    pprint(related_models_mapping)

    print("Cache cleaners")
    pprint(cache_cleaners)
